from .dimensions import SIDimensions, dim_div, dim_mul, dim_pow, dim_str
from .util import immutable, flyweight


@flyweight
@immutable
class SIUnit:
    """SIUnit is a scaling of SI base unit dimensions.

    :cvar factory:
    :vartype factory: ``Callable[[_T, SIUnit], _Q]``

    :param scale: The scaling factor of base SI dimensions.
    :type scale: ``numbers.Real``
    :param dimensions: The base SI dimensions.
    :type dimensions: ``tuple``
    """

    __slots__ = ("scale", "dimensions", "__weakref__")

    #:
    #: The factory function which unit instances use to create quantities.
    #:
    #:    .. note::
    #:
    #:        SIUnit.factory is mapped to :class:`~siquant.quantities.Quantity`
    #:        in __init__ by default. However, it is *not* required to be a type,
    #:        and can be overwritten in client configuration.
    #:
    #:        It's purpose is to provide a consistent way to wrap values, and allow
    #:        simple extensibility.
    #:
    #:        :func:`~siquant.quantities.make` is just a wrapper around this factory.
    #:
    #:        .. code-block:: python
    #:
    #:            def ext_factory(q, u):
    #:                if isinstance(q, Vector):
    #:                    return VectorQuantity(q, u)
    #:                return Quantity(q, u)
    #:
    #:            SIUnit.factory = staticmethod(ext_factory)
    #:
    factory = None

    @staticmethod
    def Unit(scale=1.0, kg=0, m=0, s=0, k=0, a=0, mol=0, cd=0):
        """Create a new SIUnit with a scale of provided base units.

        :param scale: The linear scaling factor of base units.
        :type scale: ``numbers.Real``
        :param kg: The exponent of kilograms.
        :type kg: ``numbers.Real``
        :param m: The exponent of meters.
        :type m: ``numbers.Real``
        :param s: The exponent of seconds.
        :type s: ``numbers.Real``
        :param k: The exponent of degrees Kelvin.
        :type k: ``numbers.Real``
        :param a: The exponent of Amperes.
        :type a: ``numbers.Real``
        :param mol: The exponent of mols.
        :type mol: ``numbers.Real``
        :param cd: The exponent of candelas.
        :type cd: ``numbers.Real``
        :rtype: :class:`SIUnit`
        """
        return SIUnit(scale, SIDimensions(kg=kg, m=m, s=s, k=k, a=a, mol=mol, cd=cd))

    def __init__(self, scale, dimensions):
        if scale <= 0:
            raise ValueError("SIunit scale must be positive.")
        super().__setattr__("scale", scale)
        super().__setattr__("dimensions", dimensions)

    kg = property(lambda self: self.dimensions[0])
    m = property(lambda self: self.dimensions[1])
    s = property(lambda self: self.dimensions[2])
    k = property(lambda self: self.dimensions[3])
    a = property(lambda self: self.dimensions[4])
    mol = property(lambda self: self.dimensions[5])
    cd = property(lambda self: self.dimensions[6])

    def base_units(self):
        """Get an SIUnit of equivalent dimensions with a scaling factor of 1.0.

        :rtype: :class:`~siquant.units.SIUnit`
        """
        return SIUnit(1.0, self.dimensions)

    def compatible(self, units):
        """Check whether the provided units are of the same dimensions.

        :param units: The units to check.
        :type units: :class:`~siquant.units.SIUnit`
        :rtype: ``bool``
        """
        return self.dimensions == units.dimensions

    def __call__(self, *args):
        """Create quantities these units for all arguments.

        :param args: The values to tag with units.
        :rtype: ``_Q`` = :class:`~siquant.quantities.Quantity`
        """
        return (self.factory(value, self) for value in args)

    def __mul__(self, rhs):
        if isinstance(rhs, SIUnit):
            return SIUnit(
                self.scale * rhs.scale, dim_mul(self.dimensions, rhs.dimensions)
            )
        return self.factory(rhs, self)

    def __rmul__(self, lhs):
        return self.factory(lhs, self)

    def __truediv__(self, rhs):
        if isinstance(rhs, SIUnit):
            return SIUnit(
                self.scale / rhs.scale, dim_div(self.dimensions, rhs.dimensions)
            )
        return self.factory(1 / rhs, self)

    def __rtruediv__(self, lhs):
        return self.factory(lhs, ~self)

    def __pow__(self, rhs):
        try:
            return SIUnit(self.scale ** rhs, dim_pow(self.dimensions, rhs))
        except TypeError:
            return NotImplemented

    def __invert__(self):
        return SIUnit(1 / self.scale, dim_div(SIDimensions(), self.dimensions))

    def __eq__(self, other):
        if isinstance(other, SIUnit):
            return self.scale == other.scale and self.dimensions == other.dimensions
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, SIUnit):
            return self.scale != other.scale or self.dimensions != other.dimensions
        return NotImplemented

    def __hash__(self):
        return hash((self.scale, self.dimensions))

    def __str__(self):
        return "%g*%s" % (self.scale, dim_str(self.dimensions))

    def __repr__(self):
        return "SIUnit(%f, %r)" % (self.scale, self.dimensions)



import pytest

from siquant.units import SIUnit
from siquant.quantities import Quantity as Q


@pytest.fixture
def unit():
    return SIUnit.Unit(scale=10, kg=1, m=2, s=3, k=4, a=5, mol=6, cd=7)


def test_unit_create():
    with pytest.raises(ValueError):
        SIUnit.Unit(0, kg=1, m=2, s=3, k=4, a=5, mol=6, cd=7)
    with pytest.raises(ValueError):
        SIUnit.Unit(-1, kg=1, m=2, s=3, k=4, a=5, mol=6, cd=7)


def test_hash(unit):

    other_scale = SIUnit(1, unit.dimensions)
    assert hash(unit) != hash(other_scale)
    assert hash(unit) == hash(SIUnit.Unit(10) * other_scale)

    other_dims = SIUnit.Unit(10)
    assert hash(unit) != other_dims


def test_unit_accessors(unit):
    assert unit.scale == 10
    assert unit.kg == 1
    assert unit.m == 2
    assert unit.s == 3
    assert unit.k == 4
    assert unit.a == 5
    assert unit.mol == 6
    assert unit.cd == 7


def test_unit_cmp(unit):
    assert unit == unit
    assert unit != SIUnit.Unit()

    assert not unit == object()
    assert unit != object()


def test_unit_call(unit):
    meters = SIUnit.Unit(1, m=1)

    (a, b, c, d, e) = meters(1, 2, 3, 4, 5)
    assert 1 * meters == a
    assert 2 * meters == b
    assert 3 * meters == c
    assert 4 * meters == d
    assert 5 * meters == e


def test_unit_mul(unit):
    original_dimensions = unit.dimensions

    assert unit * SIUnit.Unit(2) == SIUnit(20.0, original_dimensions)
    assert SIUnit.Unit(2) * unit == SIUnit(20.0, original_dimensions)
    assert unit * unit == SIUnit.Unit(100.0, kg=2, m=4, s=6, k=8, a=10, mol=12, cd=14)

    assert 2 * unit == Q(2, unit)
    assert unit * 2 == Q(2, unit)

    assert 2 * unit * unit == Q(2, unit * unit)
    assert unit * (2 * unit) == Q(2, unit * unit)

    unit2 = unit
    unit2 *= unit2
    assert unit2 == unit * unit

    none_q1 = None * unit
    none_q2 = unit * None

    assert none_q1 == none_q2

    with pytest.raises(TypeError):
        none_q1 < none_q2


def test_unit_div(unit):
    original_dimensions = unit.dimensions
    inverted_unit = SIUnit.Unit(1 / 10, kg=-1, m=-2, s=-3, k=-4, a=-5, mol=-6, cd=-7)
    assert unit / SIUnit.Unit(2) == SIUnit(5, original_dimensions)
    assert SIUnit.Unit(2) / unit == SIUnit(2 / 10, inverted_unit.dimensions)
    assert unit / unit == SIUnit.Unit(1)

    assert 2 / unit == Q(2, inverted_unit)
    assert unit / 2 == Q(1 / 2, unit)

    assert unit / 2 / unit == Q(1 / 2, SIUnit.Unit(1))
    assert unit / (2 / unit) == Q(50, SIUnit.Unit(1, 2, 4, 6, 8, 10, 12, 14))

    unit /= unit
    assert unit == SIUnit.Unit(1)

    none_q = None / unit
    assert none_q.quantity is None

    with pytest.raises(TypeError):
        unit / None


def test_unit_base_units(unit):
    assert unit.base_units() == unit / SIUnit.Unit(unit.scale)


def test_unit_pow(unit):
    assert unit ** 2 == unit * unit
    assert unit ** 3 == unit * unit * unit
    assert unit ** -1 == SIUnit.Unit(1) / unit
    assert unit ** -2 == SIUnit.Unit(1) / unit / unit

    with pytest.raises(TypeError):
        unit = unit ** unit


def test_unit_invert(unit):
    assert ~unit == SIUnit.Unit(1 / 10, kg=-1, m=-2, s=-3, k=-4, a=-5, mol=-6, cd=-7)


def test_unit_compatible(unit):
    assert unit.compatible(unit)
    assert unit.compatible(SIUnit.Unit(2) * unit)
    assert not unit.compatible(unit * unit)
    assert not unit.compatible(SIUnit.Unit(1))



import math

from siquant.units import SIUnit

#:
nano = SIUnit.Unit(1 / 1e9)

#:
micro = SIUnit.Unit(1 / 1e6)

#:
milli = SIUnit.Unit(1 / 1000)

#:
centi = SIUnit.Unit(1 / 100)

#:
deci = SIUnit.Unit(1 / 10)

#:
unity = SIUnit.Unit(1)

#:
deca = SIUnit.Unit(10)

#:
hecta = SIUnit.Unit(100)

#:
kilo = SIUnit.Unit(1000)

#:
mega = SIUnit.Unit(1e6)

#:
giga = SIUnit.Unit(1e9)

#:
tera = SIUnit.Unit(1e12)

#:
meters = SIUnit.Unit(m=1)

#:
kilograms = SIUnit.Unit(kg=1)

#:
seconds = SIUnit.Unit(s=1)

#:
kelvin = SIUnit.Unit(k=1)

#:
amperes = SIUnit.Unit(a=1)

#:
mols = SIUnit.Unit(mol=1)

#:
candelas = SIUnit.Unit(cd=1)

#:
radians = SIUnit.Unit(1)

#:
steradians = SIUnit.Unit(1)

#:
degrees = SIUnit.Unit(math.pi / 180) * radians

#:
hertz = unity / seconds

#:
newtons = kilograms * meters / seconds ** 2

#:
pascals = newtons / meters ** 2

#:
joules = newtons * meters

#:
watts = joules / seconds

#:
coulombs = amperes * seconds

#:
volts = watts / amperes

#:
farads = coulombs / volts

#:
ohms = volts / amperes

#:
siemens = ~ohms

#:
webers = volts * seconds

#:
teslas = webers / meters ** 2

#:
henrys = webers / amperes

#:
lumens = candelas * steradians

#:
lux = lumens / meters ** 2

#:
becquerels = ~seconds

#:
sieverts = grays = joules / kilograms

#:
katals = mols / seconds

#:
nanometers = nano * meters

#:
micrometers = micro * meters

#:
millimeters = milli * meters

#:
centimeters = centi * meters

#:
decimeters = deci * meters

#:
decameters = deca * meters

#:
hectameters = hecta * meters

#:
kilometers = kilo * meters

#:
liters = SIUnit.Unit(1000) * centimeters ** 3

#:
milliliters = milli * liters

#:
microliters = micro * liters

#:
grams = kilograms / kilo

#:
milligrams = milli * grams

#:
micrograms = micro * grams

#:
tonnes = SIUnit.Unit(1000) * kilograms

#:
kilotonnes = kilo * tonnes

#:
nanoseconds = nano * seconds

#:
microseconds = micro * seconds

#:
milliseconds = milli * seconds

#:
minutes = SIUnit.Unit(60) * seconds

#:
hours = SIUnit.Unit(60) * minutes

#:
days = SIUnit.Unit(24) * hours

#:
weeks = SIUnit.Unit(7) * days

#:
years = SIUnit.Unit(365) * days

#:
gals = centimeters / seconds ** 2

#:
g_0 = SIUnit.Unit(9.80665) * meters / seconds ** 2

#:
micropascals = micro * pascals

#:
millipascals = milli * pascals

#:
kilopascals = kilo * pascals

#:
megapascals = mega * pascals

#:
gigapascals = giga * pascals

#:
millijoules = milli * joules

#:
kilojoules = kilo * joules

#:
megajoules = mega * joules

#:
gigajoules = giga * joules

#:
terajoules = tera * joules

#:
milliwatts = milli * watts

#:
kilowatts = kilo * watts

#:
megawatts = mega * watts

#:
gigawatts = giga * watts

#:
terawatts = tera * watts

#:
millivolts = milli * volts

#:
kilovolts = kilo * volts

#:
megavolts = mega * volts

#:
gigavolts = giga * volts

#:
teravolts = tera * volts

#:
micronewtons = micro * newtons

#:
millinewtons = milli * newtons

#:
kilonewtons = kilo * newtons

#:
newton_meters = newtons * meters

#:
kilonewton_meters = kilo * newton_meters

#:
meters_per_second = meters / seconds

# etc ...



try:
    from operator import matmul
except ImportError:
    pass  # matmul not supported

from copy import copy, deepcopy
from functools import total_ordering

from .exceptions import UnitMismatchError, unexpected_type_error
from .util import immutable
from .unitless import Unitless


def make(quantity, units):
    """Entry point for creating quantities consistently

    :param quantity: The value to tag with specific units.
    :type quantity: ``_T``
    :param units: The units of this quantity.
    :type units: :class:`~siquant.units.SIUnit`
    :rtype: ``_Q`` = :class:`~Quantity`
    """
    return units.factory(quantity, units)


def converter(units):
    """Create a converter function which will return Quantities.

    :param units: The units to convert a quantity to.
    :type units: :class:`~siquant.units.SIUnit`
    :rtype: ``Callable[[Any], _Q]``
    """

    def _converter(value):
        if isinstance(value, Quantity):
            return value.cvt_to(units)
        return make(value, units)

    return _converter


def validator(dimensions):
    """Create a validator function which checks if a value matches expected dimensions.

    .. seealso::

        Predefined :mod:`~siquant.dimensions`.

        Creating new :func:`~siquant.dimensions.SIDimensions`.

    :param dimensions: The expected dimensions.
    :type dimensions: ``tuple``
    :rtype: ``Callable[[Any, ...], bool]``
    """

    if not isinstance(dimensions, tuple):
        raise unexpected_type_error("dimensions", tuple, dimensions)
    if not len(dimensions) == 7:
        raise ValueError("Dimensions tuple must have 7 elements.", dimensions)

    def _validator(*values):
        return are_of(dimensions, *values)

    return _validator


def are_of(dimensions, *quantities):
    """Check if quantities all match dimensions.

    :param dimensions: The expected dimensionality.
    :type dimensions: ``tuple``
    :param quantities: Variadic. The instances to check against.
    :rtype: ``bool``
    """
    try:
        return all(q.is_of(dimensions) for q in quantities)
    except AttributeError:
        return False


@immutable
@total_ordering
class Quantity:
    """Quantity wraps a value with units and provides arithmetic passthrough operations.

    .. note::

        Creation of Quantity directly is discouraged.

        The preferred method are:

        q = value * si.meters

        q = make(value, si.meters)

        Both of these methods delegate instantiation to
        :attr:`~siquant.units.SIUnit.factory` in order to
        more easily support clean extensibility.

    :ivar quantity: The wrapped value. read only.
    :vartype quantity: ``_T``
    :ivar units: The units of this quantity. read only.
    :vartype units: :class:`~siquant.units.SIUnit`

    :param quantity: The quantity to be wrapped.
    :type quantity: ``_T``
    :param units: The units the quantity's value is expressed in.
    :type units: :class:`~siquant.units.SIUnit`
    """

    __slots__ = ("quantity", "units")

    def __init__(self, quantity, units):
        if isinstance(quantity, Quantity):
            units = quantity.units * units
            quantity = quantity.quantity
        super().__setattr__("quantity", quantity)
        super().__setattr__("units", units)

    def is_of(self, dimensions):
        """

        :param dimensions:
        :return: ``bool``
        """
        return self.units.dimensions == dimensions

    def get_as(self, units):
        """

        :param units: The units to express the underlying value in.
        :type units: :class:`~siquant.units.SIUnit`
        :rtype: ``_T``
        """
        if self.units == units:
            return self.quantity
        if not self.units.compatible(units):
            raise UnitMismatchError(self.units, units)
        return self.units.scale / units.scale * self.quantity

    def cvt_to(quantity, units):
        """Create an equivalent Quantity expressed in the provided units.

        :param units: The units to express this quantity in.
        :type units: :class:`~siquant.units.SIUnit`
        :return: ``_Q`` = :class:`~siquant.quantities.Quantity`
        """
        return make(quantity.get_as(units), units)

    def round_to(self, units, places=0):
        """Create an equivalent Quantity rounded in provided units.

        :param units: The units to express the quantity in.
        :type units: :class:`~siquant.units.SIUnit`
        :param places: The number of decimal places to round to.
        :type places: ``int``
        :rtype: ``_Q`` = :class:`~siquant.quantities.Quantity`
        """
        return make(round(self.get_as(units), places), units)

    def compatible(self, other):
        """

        :param other: The quantity to check for dimensional compatibility.
        :type other: ``_Q`` = :class:`~siquant.quantities.Quantity`
        :rtype: bool
        """
        if isinstance(other, Unitless):
            other = make(other, self.units.Unit())
        if not isinstance(other, Quantity):
            raise unexpected_type_error("other", Quantity, other)
        return self.units.compatible(other.units)

    def __abs_epsilon(self, atol=1e-6):
        if isinstance(atol, Quantity):
            return atol
        return atol * self.units

    def __rel_epsilon(self, other, rtol=1e-9):
        return rtol * max(abs(self), abs(other), 1 * self.units)

    def abs_approx(self, other, atol=1e-6):
        return self.approx(other, rtol=0, atol=atol)

    def rel_approx(self, other, rtol=1e-9):
        return self.approx(other, rtol=rtol, atol=0)

    def approx(self, other, rtol=1e-9, atol=1e-6):
        """

        :raises: ``TypeError`` if other is not a Quantity

        :param other:
        :param rtol:
        :type rtol:
        :param atol:
        :type atol:
        :return:
        """
        if not self.compatible(other):
            return False

        epsilon = max(self.__rel_epsilon(other, rtol), self.__abs_epsilon(atol))
        return abs(other - self) <= epsilon

    def __add__(self, other):
        if isinstance(other, Quantity):
            return make(self.quantity + other.get_as(self.units), self.units)
        if isinstance(other, Unitless):
            return self + make(other, self.units.Unit())
        return NotImplemented

    def __radd__(self, other):
        if isinstance(other, Unitless):
            return make(other, self.units.Unit()) + self
        return NotImplemented

    __iadd__ = __add__

    def __sub__(self, other):
        if isinstance(other, Quantity):
            return make(self.quantity - other.get_as(self.units), self.units)
        if isinstance(other, Unitless):
            return self - make(other, self.units.Unit())
        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, Unitless):
            return make(other, self.units.Unit()) - self
        return NotImplemented

    __isub__ = __sub__

    def __neg__(self):
        return make(-self.quantity, self.units)

    def __bool__(self):
        return bool(self.quantity)

    def __eq__(self, other):
        if isinstance(other, Quantity):
            return self.units.compatible(other.units) and self.quantity == other.get_as(
                self.units
            )
        if isinstance(other, Unitless):
            return self == make(other, self.units.Unit())
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Quantity):
            return not self.units.compatible(
                other.units
            ) or self.quantity != other.get_as(self.units)
        if isinstance(other, Unitless):
            return self != make(other, self.units.Unit())
        return NotImplemented

    def __hash__(self):
        return hash((self.quantity * self.units.scale, self.units.base_units()))

    def __str__(self):
        return "%s %s" % (self.quantity, self.units)

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.quantity, self.units)

    def __abs__(self):
        return make(abs(self.quantity), self.units)

    def __lt__(self, other):
        if isinstance(other, Quantity):
            return self.quantity < other.get_as(self.units)
        return NotImplemented

    def __mul__(self, rhs):
        if isinstance(rhs, Quantity):
            return make(self.quantity * rhs.quantity, self.units * rhs.units)
        return make(self.quantity * rhs, self.units)

    __imul__ = __mul__

    def __rmul__(self, lhs):
        return make(lhs * self.quantity, self.units)

    def __matmul__(self, rhs):
        if isinstance(rhs, Quantity):
            return make(matmul(self.quantity, rhs.quantity), self.units * rhs.units)
        return make(matmul(self.quantity, rhs), self.units)

    __imatmul__ = __matmul__

    def __rmatmul__(self, lhs):
        return make(matmul(lhs, self.quantity), self.units)

    def __truediv__(self, rhs):
        if isinstance(rhs, Quantity):
            return make(self.quantity / rhs.quantity, self.units / rhs.units)
        return make(self.quantity / rhs, self.units)

    __itruediv__ = __truediv__

    def __rtruediv__(self, lhs):
        return Quantity(lhs / self.quantity, ~self.units)

    def __pow__(self, exponent):
        try:
            return make(self.quantity ** exponent, self.units ** exponent)
        except TypeError:
            return NotImplemented

    def __invert__(self):
        return make(1 / self.quantity, ~self.units)

    def __float__(self):
        return float(self.quantity)

    def __int__(self):
        return int(self.quantity)

    def __len__(self):
        return len(self.quantity)

    def __iter__(self):
        return (make(q, self.units) for q in self.quantity)

    def __getitem__(self, idx):
        return make(self.quantity[idx], self.units)

    def __copy__(self):
        return make(copy(self.quantity), self.units)

    def __deepcopy__(self, memodict):
        return make(deepcopy(self.quantity), self.units)



import weakref

from .exceptions import ImmutabilityError


def __si_immutable_setattr(inst, key, value):
    raise ImmutabilityError(inst, key)


def __si_immutable_delattr(inst, key):
    raise ImmutabilityError(inst, key)


def immutable(cls):
    cls.__setattr__ = __si_immutable_setattr
    cls.__delattr__ = __si_immutable_delattr
    return cls


def flyweight(cls):
    instances = weakref.WeakValueDictionary()
    cls.__new__ = lambda cls, *args: instances.setdefault(args, object.__new__(cls))
    return cls



"""

"""


class UnitMismatchError(ValueError):
    def __init__(self, u1, u2):
        super().__init__("Unit Mismatch: %s, %s" % (u1, u2), u1, u2)


class ImmutabilityError(AttributeError):
    def __init__(self, instance, name):
        super().__init__(
            "Can't mutate attribute '{name}'.".format(name=name), instance, name
        )


def unexpected_type_error(arg_name, expected_type, actual_value):
    return TypeError(
        "'{name}' must be {type!r} (got {value!r} that is a "
        "{actual!r}).".format(
            name=arg_name,
            type=expected_type,
            actual=actual_value.__class__,
            value=actual_value,
        ),
        arg_name,
        expected_type,
        actual_value,
    )



from siquant.units import SIUnit

from .si import pascals

#:
inches = SIUnit.Unit(25.4 / 1000, m=1)

#:
thousandths = inches / SIUnit.Unit(1000)

#:
feet = SIUnit.Unit(12) * inches

#:
yards = SIUnit.Unit(3) * feet

#:
chains = SIUnit.Unit(22) * yards

#:
furlongs = SIUnit.Unit(10) * chains

#:
miles = SIUnit.Unit(8) * furlongs

#:
acres = SIUnit.Unit(1 / 640) * miles ** 2

#:
roods = SIUnit.Unit(1 / 4) * acres

#:
perches = SIUnit.Unit(1 / 40) * roods

#:
fluid_ounces = SIUnit.Unit(1.7339) * inches ** 3

#:
gills = SIUnit.Unit(5) * fluid_ounces

#:
pints = SIUnit.Unit(20) * fluid_ounces

#:
quarts = SIUnit.Unit(40) * fluid_ounces

#:
gallons = SIUnit.Unit(160) * fluid_ounces

#:
pounds = SIUnit.Unit(0.45359237, kg=1)

#:
ounces = pounds / SIUnit.Unit(16)

#:
stones = SIUnit.Unit(14) * pounds

#:
tons = SIUnit.Unit(2240) * pounds

#:
psi = SIUnit.Unit(6894.75729) * pascals

#:
ksi = SIUnit.Unit(1000) * psi

# todo: other imperial units



import sys
import math

from copy import copy, deepcopy

import pytest
import numpy as np

import siquant.dimensions as dims

from siquant.exceptions import UnitMismatchError, ImmutabilityError
from siquant import make, converter, validator, are_of, si, SIUnit


def test_q_copying():
    a = 100 * si.meters
    b = copy(a)
    c = deepcopy(a)

    assert a is not b
    assert a is not c
    assert a == b
    assert a == c

    class Ref:
        def __init__(self, value):
            self.value = value

        def __eq__(self, other):
            return self.value == other.value

    ref = Ref(100)
    ref_ref = Ref(Ref(100))

    rq = make(ref, si.meters)
    rc = copy(rq)

    assert rq == rc
    assert rq is not rc

    rqq = make(ref_ref, si.meters)
    rqc = copy(rqq)
    rqd = deepcopy(rqq)

    assert rqq is not rqc
    assert rqq == rqc

    ref_ref.value.value = 50
    assert rqq == rqc
    assert rqq != rqd


def test_q_immutability():
    mass = 100 * si.kilograms
    with pytest.raises(ImmutabilityError):
        mass.quantity = 10
    with pytest.raises(ImmutabilityError):
        del mass.quantity


def test_q_np():
    distances = make(np.array([1, 2, 3, 4, 5, 6, 7]), si.meters)
    distances *= 2

    value = distances.get_as(si.meters)
    assert np.array_equal(value, np.array([2, 4, 6, 8, 10, 12, 14]))

    assert distances[0].approx(2 * si.meters)
    assert distances[6].approx(14 * si.meters)

    assert min(distances).approx(2 * si.meters)
    assert max(distances).approx(14 * si.meters)


def test_q_round_to():
    distance = 1.23456789 * si.meters

    assert 1235 == distance.round_to(si.millimeters).quantity

    assert 1234.6 == distance.round_to(si.millimeters, 1).quantity


def test_q_approx():

    dist1 = 1.234567890 * si.meters
    dist2 = 1.234567 * si.meters

    assert dist1.approx(dist2)
    assert dist1.approx(dist2, atol=1 * si.millimeters)

    assert dist1.abs_approx(dist2)

    with pytest.raises(UnitMismatchError):
        dist1.approx(dist2, atol=1 * si.milligrams)

    assert not dist1.approx(1.23456789 * si.kilograms)

    dist1 = 1234567890 * si.meters
    dist2 = 1234567891 * si.meters

    assert dist1.approx(dist2)
    assert dist1.rel_approx(dist2)
    assert not dist1.approx(dist2, rtol=0)
    assert not dist1.abs_approx(dist2)

    assert dist1.approx(dist2, rtol=0, atol=1 * si.meters)


@pytest.mark.skipif(
    sys.version_info < (3, 5), reason="operators.matmul introduced in py35"
)
def test_q_matmul():
    from operator import matmul

    distances = make(np.array([1, 2, 3]), si.meters)
    value = matmul(distances, distances)
    assert value.get_as(si.meters ** 2) == 14

    value = matmul(distances, np.array([1, 2, 3]))
    assert value.get_as(si.meters ** 1) == 14

    with pytest.raises(TypeError):
        # numpy really should just return NotImplemented instead of throwing a TypeError
        value = matmul(np.array([1, 2, 3]), distances)
        assert value.get_as(si.meters ** 1) == 14


def test_are_of_validator():
    dim_check = validator(dims.speed_t)

    assert dim_check(1 * si.meters / si.seconds)
    assert dim_check(1 * si.meters / si.seconds, 2 * si.meters / si.seconds)
    assert not dim_check(1 * si.meters)
    assert not dim_check(1)
    assert not dim_check(None)

    with pytest.raises(TypeError):
        validator([1, 2, 3, 4, 5, 6, 7])

    with pytest.raises(ValueError):
        validator((1, 2, 3))


def test_are_of():
    assert are_of(dims.distance_t, 1 * si.meters, 2 * si.millimeters, 3 * si.kilometers)
    assert not are_of(dims.angle_t, 1, 2, 3)
    assert not are_of(dims.distance_t, 1 * si.meters, 1 * si.kilograms)


def test_create_and_extract():
    mass = make(100, SIUnit.Unit(kg=1))
    acceleration = make(1, SIUnit.Unit(m=1, s=-2))

    force = 100 * si.newtons

    assert force.quantity == 100
    assert force.get_as(si.kilonewtons) == 0.1
    assert force == mass * acceleration


def test_hash():
    angle1 = 2 * math.pi * si.radians
    angle2 = 360 * si.degrees

    assert hash(angle1) == hash(angle2)

    circumference = angle1 * 1 * si.meters
    assert hash(circumference) != hash(angle1)


def test_q_compatible():
    dist1 = 1 * si.meters
    dist2 = 1 * si.millimeters
    area = 1 * si.meters ** 2
    assert dist1.compatible(dist2)
    assert dist2.compatible(dist1)
    assert not dist1.compatible(area)
    assert not area.compatible(dist1)

    with pytest.raises(TypeError):
        dist = 1 * si.meters
        dist.compatible(si.meters)


def test_q_eq_ne():
    dist1 = 1 * si.meters
    dist2 = 1000 * si.millimeters
    dist3 = 1 * si.kilometers
    area = 1 * si.meters ** 2

    assert dist1 == dist2
    assert dist2 == dist1

    assert dist1 != 1
    assert dist2 != 1000
    assert area != 1

    assert dist1 != dist3
    assert dist3 != dist1

    assert not area == dist1
    assert not dist1 == area

    assert not dist1 == si.meters
    assert not si.meters == dist1

    assert area != dist1
    assert dist1 != area

    assert dist1 != si.meters
    assert si.meters != dist1

    count = 500 * si.unity

    assert dist1 != count
    assert count != dist1

    assert count == 500 * si.unity
    assert count == 500
    assert 500 == count

    percent = SIUnit.Unit(1 / 100)

    hundo = 100 * percent

    assert hundo == 1


def test_q_ordering():
    dist1 = 1 * si.meters
    dist2 = 1000 * si.millimeters
    dist3 = 1 * si.kilometers
    dist4 = 1 * si.millimeters
    area = 1 * si.meters ** 2

    assert dist1 >= dist2
    assert dist2 <= dist1
    assert dist1 <= dist2
    assert dist2 >= dist1

    assert dist3 >= dist1
    assert dist1 <= dist3
    assert dist3 > dist1
    assert dist1 < dist3

    assert dist4 <= dist1
    assert dist1 >= dist4
    assert dist4 < dist1
    assert dist1 > dist4

    with pytest.raises(UnitMismatchError):
        dist1 > area
    with pytest.raises(UnitMismatchError):
        dist1 >= area
    with pytest.raises(UnitMismatchError):
        dist1 < area
    with pytest.raises(UnitMismatchError):
        dist1 <= area

    with pytest.raises(TypeError):
        dist1 > 1
    with pytest.raises(TypeError):
        dist1 >= 1
    with pytest.raises(TypeError):
        dist1 < 1
    with pytest.raises(TypeError):
        dist1 <= 1


def test_q_float():
    assert pytest.approx(1) == math.cos(0 * si.radians)
    assert pytest.approx(-1) == math.cos(math.pi * si.radians)
    assert pytest.approx(0) == math.cos(math.pi / 2 * si.radians)
    assert pytest.approx(0) == math.cos(math.pi * 3 / 2 * si.radians)


def test_q_truth():
    dist_true = 1 * si.meters
    dist_false = 0 * si.meters
    assert dist_true
    assert not dist_false


def test_q_abs():
    assert abs(1 * si.meters) == 1 * si.meters
    assert abs(-1 * si.meters) == 1 * si.meters


def test_q_invert():
    dist = 1000 * si.millimeters
    assert ~dist == 1 / 1000 / si.millimeters


def test_q_neg():
    dist = 1000 * si.millimeters
    assert -dist == -1000 * si.millimeters


def test_q_pow():
    dist = 1000 * si.millimeters
    assert dist ** 2 == 1 * si.meters ** 2
    assert dist ** -1 == 1 * si.meters ** -1

    dist = 100 * si.millimeters ** 2
    assert dist ** 0.5 == 10 * si.millimeters

    with pytest.raises(TypeError):
        dist ** dist
    with pytest.raises(TypeError):
        dist ** si.meters


def test_q_mul():
    dist = 1000 * si.millimeters
    assert 2 * dist == 2000 * si.millimeters
    assert dist * 2 == 2000 * si.millimeters
    assert dist * dist == 1000 * 1000 * si.millimeters ** 2

    dist *= 2
    assert dist == 2000 * si.millimeters


def test_q_div():
    dist = 1000 * si.millimeters
    assert 2 / dist == 1 / 500 / si.millimeters
    assert dist / 2 == 500 * si.millimeters
    assert dist / dist == 1 * si.unity

    assert dist / si.millimeters == 1000 * si.unity
    assert si.millimeters / dist == 1 / 1000 * si.unity

    with pytest.raises(TypeError):
        dist / (1, 2)
    with pytest.raises(TypeError):
        (1, 2) / dist


def test_dimensionality():
    dist = 1000 * si.millimeters
    assert dist.is_of(dims.distance_t)
    assert not dist.is_of(dims.area_t)


def test_q_add():
    dist1 = 1 * si.meters
    dist2 = 1000 * si.millimeters
    area = dist1 * dist2

    diff1 = dist1 + dist2
    diff2 = dist2 + dist1

    assert diff1 == 2 * si.meters
    assert diff1.units == si.meters

    assert diff2 == 2 * si.meters
    assert diff2.units == si.millimeters

    with pytest.raises(TypeError):
        dist1 + si.meters
    with pytest.raises(TypeError):
        si.meters + dist1

    with pytest.raises(UnitMismatchError):
        dist1 + area
    with pytest.raises(UnitMismatchError):
        area + dist1

    dist1 += dist2
    assert dist1 == 2 * si.meters

    count = 100 * si.unity
    assert count == 100
    assert count + 100 == 200


def test_q_sub():
    dist1 = 1 * si.meters
    dist2 = 1000 * si.millimeters
    area = dist1 * dist2

    diff1 = dist1 - dist2
    diff2 = dist2 - dist1

    assert diff1 == 0 * si.meters
    assert diff1.units == si.meters

    assert diff2 == 0 * si.meters
    assert diff2.units == si.millimeters

    with pytest.raises(TypeError):
        dist1 - si.meters
    with pytest.raises(TypeError):
        si.meters - dist1

    with pytest.raises(UnitMismatchError):
        dist1 - area
    with pytest.raises(UnitMismatchError):
        area - dist1

    dist1 -= dist2
    assert dist1 == 0 * si.meters

    count = 500 * si.unity
    assert count == 500 * si.unity
    assert count - 100 == 400 * si.unity
    assert count - 100 == 400
    assert 1000 - count == 500 * si.unity
    assert 1000 - count == 500


def test_q_converters():
    cvtr = converter(si.meters)
    assert cvtr(1000) == 1000 * si.meters
    assert cvtr(1000 * si.meters) == 1000 * si.meters
    assert cvtr(1000 * si.millimeters) == 1 * si.meters

    with pytest.raises(UnitMismatchError):
        cvtr(1000 * si.kilograms)



abbreviations = ("kg", "m", "s", "k", "a", "mol", "cd")


def SIDimensions(kg=0, m=0, s=0, k=0, a=0, mol=0, cd=0):
    """Create a dimensionality tuple with base si units of provided exponents.

    :param kg: The exponent of Kilograms.
    :type kg: ``numbers.Real``
    :param m: The exponent of Meters.
    :type m: ``numbers.Real``
    :param s: The exponent of Seconds.
    :type s: ``numbers.Real``
    :param k: The exponent of Kelvin.
    :type k: ``numbers.Real``
    :param a: The exponent of Amperes.
    :type a: ``numbers.Real``
    :param mol: The exponent of Mols.
    :type mol: ``numbers.Real``
    :param cd: The exponent of Candelas.
    :type cd: ``numbers.Real``
    :rtype: ``tuple``
    """
    return (kg, m, s, k, a, mol, cd)


def dim_mul(dims1, dims2):
    """Create a new dimensionality for the multiplication of dims1 by dims2.

    :param dims1: Numerator dimensions.
    :type dims1: ``tuple``
    :param dims2: Other numerator dimensions.
    :type dims2: ``tuple``
    :rtype: ``tuple``
    """
    return (
        dims1[0] + dims2[0],
        dims1[1] + dims2[1],
        dims1[2] + dims2[2],
        dims1[3] + dims2[3],
        dims1[4] + dims2[4],
        dims1[5] + dims2[5],
        dims1[6] + dims2[6],
    )


def dim_div(dims1, dims2):
    """Create a new dimensionality tuple for the division of dims1 by dims2.

    :param dims1: The numerator dimensions.
    :type dims1: ``tuple``
    :param dims2: The divisor dimensions.
    :type dims2: ``tuple``
    :rtype: ``tuple``
    """
    return (
        dims1[0] - dims2[0],
        dims1[1] - dims2[1],
        dims1[2] - dims2[2],
        dims1[3] - dims2[3],
        dims1[4] - dims2[4],
        dims1[5] - dims2[5],
        dims1[6] - dims2[6],
    )


def dim_pow(dims, exp):
    """Create a new dimensionality tuple for the exponentiation of dims by exp.

    :param dims: The base dimensions.
    :type dims: ``tuple``
    :param exp: The power to raise the base dimensions to.
    :type exp: ``numbers.Real``
    :rtype: ``tuple``
    """
    return (
        dims[0] * exp,
        dims[1] * exp,
        dims[2] * exp,
        dims[3] * exp,
        dims[4] * exp,
        dims[5] * exp,
        dims[6] * exp,
    )


def dim_str(dims):
    """Express dimensions as a human readable string.

    :param dims: The base unit dimensions to present.
    :type dims: ``tuple``
    :rtype: ``str``
    """
    return "*".join(
        "%s**%g" % (unit_abbreviation, power)
        for unit_abbreviation, power in zip(abbreviations, dims)
        if power != 0
    )


#:
angle_t = SIDimensions()

#:
solid_angle_t = SIDimensions()

#:
strain_t = SIDimensions()

#:
ratio_t = SIDimensions()

#:
mass_t = SIDimensions(kg=1)

#:
distance_t = SIDimensions(m=1)

#:
time_t = SIDimensions(s=1)

#:
temperature_t = SIDimensions(k=1)

#:
current_t = SIDimensions(a=1)

#:
amount_t = SIDimensions(mol=1)

#:
luminosity_t = SIDimensions(cd=1)

#:
frequency_t = SIDimensions(s=-1)

#:
speed_t = dim_div(distance_t, time_t)

#:
acceleration_t = dim_div(speed_t, time_t)

#:
jerk_t = dim_div(acceleration_t, time_t)

#:
jounce_t = dim_div(jerk_t, time_t)

#:
area_t = dim_pow(distance_t, 2)

#:
volume_t = dim_pow(distance_t, 3)

#:
density_t = dim_div(mass_t, volume_t)

#:
volumetric_flow_t = dim_div(volume_t, time_t)

#:
force_t = dim_mul(mass_t, acceleration_t)

#:
moment_t = dim_mul(force_t, distance_t)

#:
torque_t = moment_t

#:
impulse_t = dim_mul(force_t, time_t)

#:
momentum_t = impulse_t

#:
stress_t = dim_div(force_t, area_t)

#:
pressure_t = stress_t

#:
hydrostatic_pressure_t = dim_mul(density_t, dim_mul(acceleration_t, distance_t))

#:
stiffness_t = dim_div(force_t, distance_t)

#:
surface_tension_t = stiffness_t

#:
energy_t = dim_mul(force_t, distance_t)

#:
work_t = energy_t

#:
heat_t = energy_t

#:
power_t = dim_div(energy_t, time_t)

#:
charge_t = dim_mul(current_t, time_t)

#:
potential_t = dim_div(energy_t, current_t)

#:
capacitance_t = dim_div(charge_t, potential_t)

#:
resistance_t = dim_div(potential_t, current_t)

#:
impedance_t = resistance_t

#:
reactance_t = resistance_t

#:
conductance_t = dim_div(current_t, potential_t)

#:
magnetic_flux_t = dim_div(energy_t, current_t)

#:
magnetic_flux_density_t = dim_div(magnetic_flux_t, area_t)

#:
inductance_t = dim_mul(impedance_t, time_t)

#:
luminous_flux_t = dim_mul(luminosity_t, solid_angle_t)

#:
illuminance_t = dim_div(luminous_flux_t, area_t)

#:
molarity_t = dim_div(amount_t, volume_t)

#:
molality_t = dim_div(amount_t, mass_t)

#:
molar_mass_t = dim_div(mass_t, amount_t)

#:
entropy_t = dim_div(energy_t, temperature_t)

#:
heat_capacity_t = entropy_t

#:
specific_entropy_t = dim_div(entropy_t, mass_t)

#:
specific_heat_capacity_t = specific_entropy_t

#:
temperature_gradient_t = dim_div(temperature_t, distance_t)



from .units import SIUnit
from .quantities import Quantity, are_of, converter, validator, make
from .systems import si, imperial

SIUnit.factory = Quantity

__all__ = (
    "Quantity",
    "SIUnit",
    "si",
    "imperial",
    "are_of",
    "converter",
    "validator",
    "make",
)



import siquant.dimensions as d


def test_create():
    dims = d.SIDimensions(kg=1, m=2, s=3, k=4, a=5, mol=6, cd=7)
    assert dims == (1, 2, 3, 4, 5, 6, 7)


def test_dim_mul():
    m = d.SIDimensions(kg=1)
    a = d.SIDimensions(m=1, s=-2)
    f = d.dim_mul(m, a)
    assert f == d.SIDimensions(kg=1, m=1, s=-2)

    dist = d.SIDimensions(m=1)
    torque = d.dim_mul(f, dist)
    assert torque == d.SIDimensions(kg=1, m=2, s=-2)


def test_dim_div():
    a = d.SIDimensions(m=1, s=-2)
    f = d.SIDimensions(kg=1, m=1, s=-2)
    m = d.dim_div(f, a)
    assert m == d.SIDimensions(kg=1)

    assert d.dim_div(m, m) == d.SIDimensions()


def test_dim_pow():
    dist = d.SIDimensions(m=1)
    area = d.dim_pow(dist, 2)
    assert area == d.SIDimensions(m=2)

    sqrt = d.dim_pow(area, 0.5)
    assert sqrt == d.SIDimensions(m=1)

    v = d.dim_pow(dist, 3)
    assert v == d.SIDimensions(m=3)


def test_dim_to_str():
    s = d.dim_str(d.SIDimensions(kg=1, m=1, s=-2))
    assert s == "kg**1*m**1*s**-2"



import os.path as path

from setuptools import find_packages, setup


def readme():
    filepath = path.join(path.dirname(path.realpath(__file__)), "README.rst")
    with open(filepath) as f:
        return f.read()


PROJECT_URLS = {
    "Documentation": "https://siquant.readthedocs.io/",
    "Bug Tracker": "https://github.com/keystonetowersystems/siquant/issues",
    "Source Code": "https://github.com/keystonetowersystems/siquant",
}
KEYWORDS = ["SI", "units", "dimensional analysis", "quantities"]


INSTALL_REQUIRES = []
EXTRAS_REQUIRE = {"docs": ["sphinx", "numpy"], "tests": ["coverage", "pytest", "numpy"]}
EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["docs"] + EXTRAS_REQUIRE["tests"] + ["pre-commit"]
)


setup(
    name="siquant",
    version="4.0.0b11",
    description="Dimensional Analysis and Unit Tracking Library",
    long_description=readme(),
    url="https://github.com/keystonetowersystems/siquant",
    project_urls=PROJECT_URLS,
    keywords=KEYWORDS,
    author="Greg Echelberger",
    author_email="greg@keystonetowersystems.com",
    packages=find_packages(exclude=("tests",)),
    install_requires=INSTALL_REQUIRES,
    setup_requires=["tox"],
    extras_require=EXTRAS_REQUIRE,
    zip_safe=True,
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
)
